using ProgressMeter
using Roots
using HDF5
using DelimitedFiles
using OrderedCollections
using DataFrames
import Dates: now, format

include("estimate_statistics.jl")

function gather_coupling_data(h5grp)
    
    modes = keys(h5grp)
    ηtot = zeros(size(h5grp[modes[1]])) # assume same length for all modes

    for m ∈ modes
        @. ηtot += h5grp[m]
    end

    @. -10log10(ηtot)
end


function analyze_coupling_data(fn; Lrange=(0, 20), probability=1e-2)

    data = h5open(fn, "r")
    fibers = keys(data)
    filter!(f -> f != "RXFields", fibers)
    # make sure, the fibers are sorted by their number of modes
    sort!(fibers, by=f -> length(keys(data[f]["Efficiencies"])))

    Ltot = OrderedDict{String, Vector{Float64}}()
    ccdf = OrderedDict{String, Function}()
    thresh = OrderedDict{String, Float64}()

    for f ∈ fibers
        h5grp = data[f]["Efficiencies"]
        push!(Ltot, f => gather_coupling_data(h5grp))
        fun = ecdf(Ltot[f])
        push!(ccdf, f => x -> 1 - fun(x))
        push!(thresh, f => find_zero(x -> fun(x) - 1 + probability, Lrange))
    end

    meta = Dict(string(a) => read_attribute(data, a) for a ∈ keys(attributes(data)))
    
    close(data)

    ccdf, thresh, Ltot, meta
end


function batch_analyze_coupling_data(dirs...; xvalue="Cn2", conditions=(;), dump::Bool=false, kwargs...)
   
    fns = filter(x -> last(splitext(x)) == ".hdf5", [fn for d ∈ dirs for fn ∈ readdir(d; join=true)])
    Nf = length(fns)

    prg = Progress(Nf, desc="Evaluating datasets...", color=:yellow)

    meta = Vector{OrderedDict}(undef, Nf)
    thresh = Vector{OrderedDict}(undef, Nf)

    @showprogress for (k, fn) ∈ enumerate(fns)
        _, thresh[k], _, meta[k] = analyze_coupling_data(fn; kwargs...)
        next!(prg)
    end

    idx = [haskey(d, string(k)) ? d[string(k)] !== v : false for d ∈ meta for (k, v) ∈ pairs(conditions)]
    deleteat!(meta, idx)
    deleteat!(thresh, idx)
    deleteat!(fns, idx)

    commonkeys = intersect(keys.(meta)...)
    meta = [OrderedDict(k => m[k] for k ∈ commonkeys) for m ∈ meta]

    mdf = DataFrame(pairs(d) for d ∈ meta)
    tdf = DataFrame(pairs(d) for d ∈ thresh)
    ddf = DataFrame(:Directory => map(x -> splitdir(x)[1], fns))
    df = hcat(mdf, tdf, ddf)

    traces = [Config(type="scatter", x=gdf[!, xvalue], y=gdf[!, f], mode="markers", name=string(f)) for (k, gdf) ∈ enumerate(groupby(df, :Directory)) for f ∈ names(tdf)]
    preview(Plot(traces, Config(xaxis=Config(title=xvalue), yaxis=Config(title="Coupling Loss Threshold [dB]"))), reuse=false) |> display

    if dump
        fn = "$(format(now(), "YYYY-mm-dd-HH-MM-SS"))_coupling_analysis_$xvalue"
        # hdf5 export
        h5open(fn * ".hdf5", "w") do fid
            [fid[string(k)] = v for (k,v) ∈ pairs(eachcol(df))]
        end
        # csv export
        writedlm(fn * ".csv", Iterators.flatten(([names(df)], eachrow(df))), ';')
    end

    return df
end