using ProgressMeter
using FFTW
using Random
using DataFrames
using HDF5
using Colors
using FileIO
import Dates: now, format


include("estimate_statistics.jl")
include("ft_phase_screen.jl")
include("plot_basics.jl")

function plot_screen(; Nx=256, Ny=Nx, dx=1e-1 / Nx, dy=dx, Nsh=0, uniform=true, tol=1e-5, seed=1, wrap=true, dump=false, Cn²=1e-14, kwargs...)
    # generate phase screen
    l0 = 1e-3
    L0 = 1e1
    k0 = 2π / 1.55e-6
    z = 1e3
    r0 = (0.423 * k0^2 * z * Cn²)^(-3 / 5)
    @info "d/r0:" dx * Nx / r0

    Random.seed!(seed)

    _screen = zeros(ComplexF64, Nx, Ny)

    if uniform
        ft_phase_screen!(_screen; r0, dx, dy, l0, L0, Nsh, centered=true, kwargs...)
    else
        # have a look at phase screens with sparse (polar) sampling, might fulfill statistics (on average) but full of artifacts
        non_uniform_phase_screen!(_screen; r0, dx, dy, l0, L0, tol, centered=true, kwargs...)
    end

    screen = real(_screen)

    wrap && (@. screen = (mod2pi(screen + π) - π))

    x = fftshift(fftfreq(Nx, Nx * dx))

    if dump
        img = @. round(UInt8, mod2pi(screen + π) * 255 / 2π)
        save("$(format(now(), "YYYY-mm-dd-HH-MM-SS"))_ps_Dr0_$(round(dx * Nx / r0, sigdigits=2)).png", img)
    else
        preview(Plot(Config(type="heatmap", x=x, y=x, z=screen, zmid=0, colorscale="Greys"), Config(xaxis=Config(title="x"), yaxis=Config(title="y", scaleanchor="x"))), reuse=false) |> display
    end

    nothing
end


function phase_screen_sf(iters=100; Nx=128, Nκ=Nx, Nϕ=Nx, Nsh=7, q=3, sf_len=Nx, showplot=true, kth_sh::Int=max(Nsh, 1), dump=false, d=1e-1, polar=false, chunk=1, dtype=Float64, model=andrews)

    # generate phase screen
    l0 = 1e-3
    L0 = 1e1
    k0 = 2π / 1.55e-6
    z = 1e3
    Cn² = 1e-14
    r0 = (0.423 * k0^2 * z * Cn²)^(-3 / 5)
    dx = d / Nx # grid spacing [m]
    tol = 1e-5

    # plan ifft without scaling
    n_fftw_threads = FFTW.get_num_threads()
    FFTW.set_num_threads(1)

    # which subharmonics to generate
    subharmonics = reverse(range(Nsh, 1, step=-kth_sh))

    # define screens to generate
    approaches = (
        "FT-Method" => ft_phase_screen!,
        # ("FT + $k SH" => ft_phase_screen! for k in subharmonics)...,
        # ("PWD + $k SH" => ft_phase_screen! for k in subharmonics)...,
        "Log. distr." => non_uniform_phase_screen!,
        "Arsinh distr." => non_uniform_phase_screen!,
        # "Sqrt distr." => non_uniform_phase_screen!,
        "Log. polar" => non_uniform_phase_screen_pol!,
        "Arsinh polar" => non_uniform_phase_screen_pol!,
    )

    # distinct function arguments
    args = Dict(
        "FT-Method" => (; model),
        ("FT + $k SH" => (; Nsh, q, model) for k in subharmonics)...,
        ("PWD + $k SH" => (; Nsh, q, pwd=true, model) for k in subharmonics)...,
        "Log. distr." => (; Nκx=Nκ, tol, fun="log", nthreads=1, model),
        "Arsinh distr." => (; Nκx=Nκ, tol, fun="asinh", nthreads=1, model),
        "Sqrt distr." => (; Nκx=Nκ, tol, fun="sqrt", nthreads=1, model),
        "Log. polar" => (; Nκ, Nϕ, tol, fun="log", nthreads=1, model),
        "Arsinh polar" => (; Nκ, Nϕ, tol, fun="asinh", nthreads=1, model),
    )

    # we use two screens (real & imaginary part) in each iteration
    iters >>= 1
    # we generate chunks of phase screens
    iters ÷= chunk

    # preallocate structure functions
    D = Dict(k => zeros(sf_len, iters) for (k, v) ∈ approaches)

    prg = Progress(iters, desc="Iterate...")
    Threads.@threads for i ∈ Base.OneTo(iters)

        # preallocating requires thread-local storage, not recommended, see also https://julialang.org/blog/2023/07/PSA-dont-use-threadid/
        screen = zeros(Complex{dtype}, Nx, Nx, chunk)

        for (k, f!) ∈ approaches
            # fix random seed --> comparison
            Random.seed!(i)
            # compute phase screens
            f!(screen; r0, dx, l0, L0, args[k]...)

            # Structure function of complex valued function = structure function of real + imaginary part
            if polar
                D[k][:, i] .= 1 / 2 .* calculate_isotropic2d_structure_function_ft(screen; sf_len, tol=1e-6)
            else
                D[k][:, i] .= 1 / 4 .* (calculate_structure_function_ft(screen; dim=1, sf_len) .+ calculate_structure_function_ft(screen; dim=2, sf_len))
            end
        end

        next!(prg)
    end

    D_mean = Dict(k => vec(mean(D[k], dims=2)) for (k, v) ∈ approaches)

    # reset FFTW threads
    FFTW.set_num_threads(n_fftw_threads)


    # compute structure function numerically (avoids difficulties in evaluating the Kummer function 1F1), make use of spectral decay
    r = (0:sf_len-1) * dx
    κmin = 0
    Φ = phase_spectrum(r0, l0, L0, model)
    D_th = num_sf(Φ, r, κmin)

    relerr(x, y) = iszero(y) ? missing : abs(1 - x / y)

    sfdata = Dict(("Theory" => D_th, (k => D_mean[k] for (k, v) ∈ approaches)...))
    errdata = Dict((k => relerr.(D_mean[k], D_th) for (k, v) ∈ approaches))

    # Theoretical sf plus FT method and all the others
    sf_traces = [Config(type="scatter", x=r, y=v, name=k) for (k, v) ∈ sfdata]
    # Plot rel. errors
    err_traces = [Config(type="scatter", x=r, y=v, name=k) for (k, v) ∈ errdata]
    # pushfirst!(err_traces, Config(type="scatter", x=r, y=fill(NaN, size(r)), showlegend=false))

    # Print RMS relative error, compare Paulson 2020
    [print("RMS rel. error of " * string(k) * ": " * string(mean(abs2, skipmissing(v)) |> sqrt) * "\n") for (k, v) ∈ errdata]

    # Plot structure functions
    p = Plot(sf_traces, Config(xaxis=Config(title="Observation point distance r [m]"), yaxis=Config(title="Dᵩ(r)", rangemode="nonnegative")))
    # Plot relative errors
    perr = Plot(err_traces, Config(xaxis=Config(title="Observation point distance r [m]"), yaxis=Config(title="Relative error δᵣ(r)", type="log"), colorway=circshift(default_colorway, -1)))


    if showplot
        display(preview(p, reuse=false))
        display(preview(perr, reuse=false))
    end

    if dump
        # fnp = "D_iters_$(iters)_SH_$(Nsh)_N_$(Nx)_Nk_$(Nκ)_d_$(d)" * ".svg"
        # fnperr = "D_iters_$(iters)_SH_$(Nsh)_N_$(Nx)_Nk_$(Nκ)_d_$(d)" * "_rel_err.svg"
        # println("\n" * fnp)
        # println(fnperr)
        # savefig(p, fnp)
        # savefig(perr, fnperr)

        sf_df = coalesce.(DataFrame(sfdata), NaN)
        err_df = coalesce.(DataFrame(errdata), NaN)

        # hdf5 output
        fn = "$(format(now(), "YYYY-mm-dd-HH-MM-SS"))_sf_iters_$(iters)_SH_$(Nsh)_N_$(Nx)_Nk_$(Nκ)_d_$(d)"
        h5open(fn * ".hdf5", "w") do fid
            create_group(fid, "SF")
            [fid["SF"][string(k)] = v for (k, v) ∈ pairs(eachcol(sf_df))]
            create_group(fid, "RelErr")
            [fid["RelErr"][string(k)] = v for (k, v) ∈ pairs(eachcol(err_df))]

            attributes(fid)["xaxis"] = collect(r)
            attributes(fid)["d"] = d
            attributes(fid)["z"] = z
            attributes(fid)["r0"] = r0
            attributes(fid)["Cn2"] = Cn²
            attributes(fid)["l0"] = l0
            attributes(fid)["L0"] = L0
            attributes(fid)["model"] = string(model)
            attributes(fid)["Nx"] = Nx
            attributes(fid)["Nk"] = Nκ
            attributes(fid)["Nphi"] = Nϕ
            attributes(fid)["Nsh"] = Nsh
            attributes(fid)["q"] = q
            attributes(fid)["tol"] = tol
        end
    end

    nothing
end