include("turbulent_channel_model.jl")

import Dates: today
import FourierTools: resample, center_extract, center_pos

# From Zheng 2016 Opt. Expr.: "The first [fiber] is a standard G.652 SMF with fundamental mode diameter of 9 μm. The second one is a 2MF with core diameter of 18 μm and fundamental mode diameter of 15.6 μm. The last one is a 4MF whose core diameter is 24 μm and fundamental mode diameter is 18.2 μm.
# see specs of Corning® SMF-28® Ultra Optical Fiber: https://www.corning.com/media/worldwide/coc/documents/Fiber/product-information-sheets/PI-1424-AEN.pdf
default_fibers = (
    SMF=(4.1e-6, 0.1, 5.2e-6),
    FMF3=(9e-6, 0.1, 7.8e-6),
    FMF6=(12e-6, 0.1, 9.1e-6),
)
default_proteus = (; (Symbol("CP$k") => (300e-6, k) for k ∈ 1:6)...)


function plot_fields(N=256; domain::Domain{T}=Domain(0.2/N, 1e1, N, 101, Inf), beam::Beam{T}=Beam(0, 0, 1.55e-6, 2e-2), waveguide=FiberModes(9e-6, 0.1, 7.8e-6), Cn²=1e-14, uniform::Bool=false, model=andrews, seed=1, n_os=1, plotrx=true, plotfocused=false, plotdest=false, pw=false) where {T}

    # compute optimal focal length
    f = optimize_focus(domain, beam, waveguide; debug=true)

    # generate modes in observation plane, focal plane resolution automatically chosen. Critical sampling for dx = beam.λ*f/domain.d
    push!(waveguide.modes, pairs(setup_modes(waveguide, domain, beam, f))...)

    # reproducible results
    Random.seed!(seed)

    # propagate and couple
    uout = turbulent_channel_model(; domain, beam, Cn², uniform, pw, model)
    urx = copy(uout)
    η = freespace_coupling!(uout, domain, beam, waveguide, f)

    r0 = (0.423 * beam.k0^2 * domain.dz * Cn²)^(-3 / 5)
    wz = beam.w0 * sqrt(1 + (domain.z[end] * beam.λ / π / beam.w0^2))
    
    @info "D/r0: " min(2domain.rA, domain.Dx, domain.Dy, 2wz) / r0
    @info "σᵣ²: " 1.23Cn² * beam.k0^(7 / 6) * domain.z[end]^(11 / 6)

    # focal plane grid and destination boundaries
    dxf, dyf = focal_plane_res(domain, beam, waveguide, f)
    x = fftfreq(domain.Nx, domain.Nx * dxf) |> fftshift
    y = fftfreq(domain.Ny, domain.Ny * dyf) |> fftshift
    ref = _magnification_ref(waveguide, beam.k0)

    # extract center and oversample:
    extract_and_oversample(z) = resample(center_extract(z, size(z) .÷ n_os), size(z))
    x0 = first(center_extract(x, size(x) .÷ n_os))
    y0 = first(center_extract(y, size(y) .÷ n_os))
    dx = step(x) / n_os
    dy = step(y) / n_os

    # reference circle, e.g., fiber core interface
    shapes = [Config(type="circle", x0=-ref, x1=ref, y0=-ref, y1=ref, line=Config(width=2, dash="dash", color="white"))]

    # plot receive field
    plotrx && display(preview(Plot(complex_plot(x0=first(domain.x), dx=step(domain.x), y0=first(domain.y), dy=step(domain.y), z=urx), Config(xaxis=Config(title="x"), yaxis=Config(title="y", scaleanchor="x"), showscale=true)), reuse=false))

    # plot focused field
    plotfocused && display(preview(Plot(complex_plot(x0=x0, dx=dx, y0=y0, dy=dy, z=extract_and_oversample(uout)), Config(xaxis=Config(title="x"), yaxis=Config(title="y", scaleanchor="x"), showscale=true; shapes)), reuse=false))

    # plot reference mode fields
    plotdest && [display(preview(Plot(complex_plot(x0=x0, dx=dx, y0=y0, dy=dy, z=extract_and_oversample(v), zmid=0), Config(xaxis=Config(title="x"), yaxis=Config(title="y", scaleanchor="x"), title=string(k); shapes)), reuse=false)) for (k, v) ∈ pairs(waveguide.modes)]

    ηtot = sum(values(η))
    @info "Total coupling loss: " -10log10(ηtot)

    η
end


function analyze_turbulent_channel_model(iters::Int=100, T::Type=Float64; waveguides=generate_wgs(; default_fibers...), Nx::Int=128, Cn²=1e-14, λ=1.55e-6, dz=5e1, Nz::Int=21, uniform::Bool=false, model::Function=andrews, dump::Bool=false, showplot::Bool=false, individual::Bool=false, fieldstats::Bool=false, pw::Bool=false, meta=(;))

    # domain and beam
    domain = Domain{T}(20e-2/Nx, dz, Nx, Nz, 5e-2)
    beam = Beam{T}(0, 0, λ, 1e-2)
    # domain = Domain{T}(6e-3/Nx, dz, Nx, Nz, Inf)
    # beam = Beam{T}(0, 0, λ, 7.9e-4) # SMF with f = 8mm collimating lens

    
    # info on turbulent channel
    l0 = T(1e-3)
    L0 = T(1e1)
    r0 = (0.423 * beam.k0^2 * domain.Dz * Cn²)^(-3 / 5)
    @info "Fried parameter: " r0
    # estimated beam waist radius at receiver
    wz = beam.w0 * sqrt(1 + (domain.Dz * beam.λ / π / beam.w0^2))
    D_r0 = min(2domain.rA, domain.Dx, domain.Dy, 2wz) / r0
    @info "Final D/r₀:" D_r0
    
    # check simulation validity
    (domain.dx / l0 > 2) && (@warn "Resolution insufficient for sampling of inner scale")
    # compare Schmidt 2010 (7.13)
    (domain.dx ≥ domain.dz/2beam.w0 * beam.λ) && (@warn "Grid resolution insufficient for free-space Fresnel propagation!")
    
    # generate coupling efficiency & focal length dicts
    η = Dict()
    f = Dict()

    @info "Set up modes..."
    @time for (wgname, wg) ∈ pairs(waveguides)
        push!(f, wgname => optimize_focus(domain, beam, wg))
        push!(wg.modes, pairs(setup_modes(wg, domain, beam, f[wgname]))...)
        push!(η, wgname => OrderedDict(k => zeros(iters) for k ∈ keys(wg.modes)))
    end

    n_fftw_threads = FFTW.get_num_threads()
    FFTW.set_num_threads(1)
    n_blas_threads = BLAS.get_num_threads()
    BLAS.set_num_threads(1)

    prg = Progress(iters, desc="Iterate...")
    fieldstats && (urx = zeros(Complex{T}, Nx, Nx, iters))
    @time Threads.@threads :static for i ∈ Base.OneTo(iters)

        Random.seed!(i)

        uout = turbulent_channel_model(; domain, beam, Cn², uniform, model, l0, L0, pw)
        fieldstats && (urx[:, :, i] .= uout)

        for (wgname, wg) ∈ pairs(waveguides)
            ηi = freespace_coupling(uout, domain, beam, wg, f[wgname])

            for k ∈ keys(wg.modes)
                @inbounds η[wgname][k][i] = ηi[k]
            end
        end

        next!(prg)
    end

    FFTW.set_num_threads(n_fftw_threads)
    BLAS.set_num_threads(n_blas_threads)

    if fieldstats  
        # scintillation index
        σ²ᵢ = estimate_scintillation_index(urx)
        
        # Plane wave, weak fluctuations unbounded inertial range (Kolmogorov)
        σ²ᵣ = 1.23Cn²*beam.k0^(7//6)*domain.z[end]^(11//6)
        @info "Theory (weak): " σ²ᵣ
        @info "Theory (strong): " exp(0.49σ²ᵣ/(1 + 1.11σ²ᵣ^(6//5))^(7//6) + 0.51σ²ᵣ/(1 + 0.69σ²ᵣ^(6//5))^(5//6)) - 1        
        
        # degree of coherence
        mcf = estimate_mcf(urx; polar=true)
        
        r = (0:length(mcf)-1) * domain.dx
        # unscaled PSD and corresponding semianalytical structure function
        D_pl = num_sf(model(1; κ0=2π / L0, κl=3.3 / l0, κm=5.92 / l0), r)
        r0_est, fit = estimate_r0(mcf, D_pl)
        @info "Estimated Fried parameter: " r0_est
        mcf_th_pl = exp.(-1//2 * 0.49r0^(-5//3) * D_pl)
        
        if showplot
            preview(Plot(Config(type="heatmap", x=domain.x, y=domain.y, z=σ²ᵢ, zmin=0, zmax=min(2mean(σ²ᵢ), 6)), Config(xaxis=Config(title="x"), yaxis=Config(title="y", scaleanchor="x"), showscale=true)), reuse=false) |> display

            preview(Plot([Config(type="scatter", x=r, y=mcf_th_pl, name="Theory (PL)"), Config(type="scatter", x=r, y=mcf, mode="markers", name="Estimate"), Config(type="scatter", x=r, y=fit, mode="markers+lines", name="Fit")], Config(xaxis=Config(title="x"), yaxis=Config(title="y"))), reuse=false) |> display
        end
    end

    # stats = estimate_pdf
    stats = estimate_ccdf

    if showplot
        traces = Config[]

        for (wgname, wg) ∈ pairs(waveguides)
            if individual
                for m ∈ keys(wg.modes)
                    x, y = stats(-10log10.(η[wgname][m]))
                    push!(traces, Config(type="scatter", x=x, y=y, name=string(wgname) * ": " * string(m), showlegend=true))
                end
            else
                x, y = stats(-10log10.(sum(values(η[wgname]))))
                push!(traces, Config(type="scatter", x=x, y=y, name=string(wgname), showlegend=true))
            end
        end
        Llim = 10(fld(max((maximum(t.x) for t ∈ traces)...), 10) + 1)

        display(preview(Plot(traces, Config(xaxis=Config(title="Coupling loss L [dB]", range=[0, Llim]), yaxis=Config(title="Outage probability", type="log", nticks=round(Int, log10(iters)) + 2))), reuse=false))
    end


    if dump
        # dump to .hdf5 file for postprocessing
        fn = "$(today())_coupling_Cn2_$(round(Cn²,sigdigits=3))_wl_$(beam.λ)_Nx_$(Nx)_iters_$(iters)_$(uniform ? "ftsh" : "nfft")"
        fieldstats && (fn * "_withrx")

        h5open(fn * ".hdf5", "w") do fid 
            for (wgname, cη)  ∈ pairs(η)
                grp = create_group(fid, string(wgname))

                attributes(grp)["focallength"] = f[wgname]

                ηgrp = create_group(grp, "Efficiencies")
                [ηgrp["M" * string(k...)] = v for (k, v) ∈ pairs(cη)]

                # save mode fields, too -> does not hurt
                mgrp = create_group(grp, "Modefields")
                [mgrp["M" * string(k...)] = v for (k, v) ∈ pairs(waveguides[wgname].modes)]
            end

            # dump fields as well if possible
            if fieldstats
                fid["RXFields", chunk=(size(urx, 1), size(urx, 2), 1), blosc=5] = urx
            end

            attributes(fid)["Dx"] = domain.Dx
            attributes(fid)["Dy"] = domain.Dy
            attributes(fid)["Dz"] = domain.Dz
            attributes(fid)["rA"] = domain.rA
            attributes(fid)["r0"] = r0
            attributes(fid)["Cn2"] = Cn²
            attributes(fid)["l0"] = l0
            attributes(fid)["L0"] = L0
            attributes(fid)["model"] = string(model)
            attributes(fid)["Nx"] = domain.Nx
            attributes(fid)["Ny"] = domain.Ny
            attributes(fid)["Nz"] = Nz
            attributes(fid)["w0"] = beam.w0
            attributes(fid)["lambda"] = beam.λ
            [attributes(fid)[string(k)] = v for (k, v) ∈ pairs(meta)]
        end
    end

    nothing
end
