using FINUFFT
using FFTW
using Bessels
using StatsBase
using Statistics
using AverageShiftedHistograms
using QuadGK
using Optim

# numerical structure / covariance function for psds of isotropic random fields
num_sf(S::Function, r, κmin=zero(eltype(r)), rtol=1e-8) = quadgk(κ -> 4π * (1 .- besselj0.(κ * r)) * S(κ) * κ, κmin, Inf, rtol=rtol) |> first
num_cov(S::Function, r, κmin=zero(eltype(r)), rtol=1e-8) = quadgk(κ -> 2π * besselj0.(κ * r) * S(κ) * κ, κmin, Inf, rtol=rtol) |> first



calculate_structure_function(x::AbstractArray{T}; dim=1, sflen=size(x, dim)) where {T} = calculate_structure_function!(zeros(real(T), sflen), x; dim)

function calculate_structure_function!(D::AbstractVector, x::AbstractArray{T,N}; dim=1) where {T,N}

    ax(k) = CartesianIndices(ntuple(d -> (d == dim) ? Base.OneTo(size(x, d) - k + 1) : axes(x, d), Val(N)))
    offset(k) = CartesianIndex(ntuple(d -> (d == dim) ? k - 1 : 0, Val(N)))

    @inbounds for k ∈ eachindex(D)
        @simd for I ∈ ax(k)
            D[k] += abs2(x[I] - x[I+offset(k)]) # magnitude squared for complex values! See Yaglom 2004 p. 88
        end
        D[k] /= length(ax(k))
    end

    D
end


function calculate_structure_function_ft(x0::AbstractArray{T,N}; dim=1, sf_len=size(x0, dim)) where {T,N}
    # calculate avg. structure function along dimension dim, compare Marcotte 1996
    fftlen = 2size(x0, dim)# - 1

    # window function/indicator matrix
    w = zeros(T, ntuple(k -> k == dim ? fftlen : 1, Val(N)))
    w[axes(x0, dim)] .= one(T)
    # zeropad to fftlen
    x = zeros(T, ntuple(k -> k == dim ? fftlen : size(x0, k), Val(N)))
    x[axes(x0)...] .= x0

    Px = plan_fft(x, dim)
    Pw = plan_fft(w, dim)

    X = Px * x
    @. x = abs2(x)
    X2 = Px * x
    W2 = Pw * w # w^2 = w

    iPx = plan_bfft(X, dim)
    iPw = plan_bfft(W2, dim)

    # compute variogram
    @. X = 2 * (real(X2 * conj(W2)) - abs2(X))
    # compute number of pairs at all lags
    @. W2 = abs2(W2)

    D0 = real((iPx * X) ./ (iPw * W2))

    idx = ntuple(k -> k == dim ? Base.OneTo(sf_len) : axes(D0, k), Val(N))
    D = vec(mean(view(D0, idx...), dims=(k < dim ? k : k + 1 for k ∈ Base.OneTo(N))))
end

function calculate_isotropic2d_structure_function_ft(x0::AbstractArray{T,N}; tol=1e-7, sf_len=size(x0, 1), nthreads=1) where {T,N}
    # calculate avg. structure function along dimension dim, compare Marcotte 1996
    fftlen = 2size(x0, 1), 2size(x0, 2)

    # window function/indicator matrix
    w = zeros(T, fftlen..., ntuple(k -> size(x0, k + 2), Val(N - 2))...)
    w[CartesianIndices(axes(x0))] .= one(T)

    r = range(0, real(T)(2π), length=fftlen[1])
    ϕ = range(0, real(T)(2π), length=fftlen[2])
    grid = ndgrid(ϕ, r)
    pol2cart!(grid...)

    # zeropad to fftlen
    x = zeros(T, fftlen..., ntuple(k -> size(x0, k + 2), Val(N - 2))...)
    x[axes(x0)...] .= x0

    P = plan_fft(x, 1:2)

    X = P * x
    @. x = abs2(x)
    X2 = P * x
    W2 = P * w # w^2 = w

    # compute variogram
    @. X = 2 * (real(X2 * conj(W2)) - abs2(X))
    # compute number of pairs at all lags
    @. W2 = abs2(W2)

    D0 = real(nufft2d2(grid..., 1, tol, X, modeord=1; nthreads) ./ nufft2d2(grid..., 1, tol, W2, modeord=1; nthreads))

    D = view(reshape(D0, fftlen..., :), axes(x0)...)
    deleteat!(mean(D, dims=(1, 3:N...)) |> vec, sf_len+1:size(D, 2))
end

function calculate_correlation_function_ft(x0::AbstractArray{T,N}; dim=1, corr_len=size(x0, dim)) where {T,N}
    fftlen = 2size(x0, dim)# - 1

    # window function
    w = zeros(T, ntuple(k -> k == dim ? fftlen : 1, Val(N)))
    w[axes(x0, dim)] .= one(T)
    # zeropad to fftlen
    x = zeros(T, ntuple(k -> k == dim ? fftlen : size(x0, k), Val(N)))
    x[axes(x0)...] .= x0

    X = fft(x, dim)
    W = fft(w, dim)
    @. X = abs2(X)
    @. W = abs2(W)

    B0 = real(bfft(X, dim) ./ bfft(W, dim))

    idx = ntuple(k -> k == dim ? Base.OneTo(corr_len) : axes(B0, k), Val(N))
    B = vec(mean(view(B0, idx...), dims=(k < dim ? k : k + 1 for k ∈ Base.OneTo(N))))
end

function calculate_isotropic2d_correlation_function_ft(x0::AbstractArray{T,N}; tol=1e-7, sf_len=size(x0, 1), nthreads=1) where {T,N}
    # calculate avg. structure function along dimension dim, compare Marcotte 1996
    fftlen = 2size(x0, 1), 2size(x0, 2)

    # window function/indicator matrix
    w = zeros(T, fftlen..., ntuple(k -> size(x0, k + 2), Val(N - 2))...)
    w[CartesianIndices(axes(x0))] .= one(T)

    r = range(0, real(T)(2π), length=fftlen[1])
    ϕ = range(0, real(T)(2π), length=fftlen[2])
    grid = ndgrid(ϕ, r)
    pol2cart!(grid...)

    # zeropad to fftlen
    x = zeros(T, fftlen..., ntuple(k -> size(x0, k + 2), Val(N - 2))...)
    x[axes(x0)...] .= x0

    X = fft(x, 1:2)
    W = fft(w, 1:2)
    @. X = abs2(X)
    @. W = abs2(W)

    B0 = real(nufft2d2(grid..., 1, tol, X, modeord=1; nthreads) ./ nufft2d2(grid..., 1, tol, W, modeord=1; nthreads))

    B = view(reshape(B0, fftlen..., :), axes(x0)...)
    deleteat!(mean(B, dims=(1, 3:N...)) |> vec, sf_len+1:size(B, 2))
end


"""
    estimate_scintillation_index(u)

    Estimate scintillation index of 3D array of complex amplitude matrices u. Extracts portion of the domain containing at least 90% of the optical power. Compare bottom of Andrews 2019 p. 48 (68).
"""
function estimate_scintillation_index(u::AbstractArray{T,3}) where {T}
    I = abs2.(u)
    dropdims(mean(abs2, I; dims=3) ./ mean(I; dims=3).^2 .- 1; dims=3)
end


"""
    estimate_mcf(u)

    Estimate mutual coherence function of 3D array of complex amplitude matrices u. Extracts portion of the domain containing at least 90% of the optical power.
"""
function estimate_mcf(u::AbstractArray{T,3}; polar::Bool=false, lq_correction=true) where {T}
    uc = copy(u)

    # identify and extract center portion of complex amplitude in 5% steps (relative to initial size)
    xstep, ystep = size(uc, 1) ÷ 20, size(uc, 2) ÷ 20

    while isapprox(sum(abs2, uc), sum(abs2, center_extract(uc, (size(uc, 1) - xstep, size(uc, 2) - ystep, size(uc, 3)))), rtol=1e-1)
        uc = center_extract(uc, (size(uc, 1) - xstep, size(uc, 2) - ystep, size(uc, 3)))
    end

    # estimate mean amplitude
    mean_uc = dropdims(sqrt.(mean(abs2, uc, dims=3)); dims=3)
    
    if lq_correction
        M, N = size(uc)
        # remove (linear) & quadratic phase, make sure we match the intensity peak!
        c = argmax(mean_uc)
        x = (1:M) .- c[1]
        y = (1:N) .- c[2]
        
        X = [x for x ∈ x, y ∈ y]
        Y = [y for x ∈ x, y ∈ y]
        
        # use overlap integral as cost function
        J(q) = begin
            # tmp = @. cis(q[1] * (X^2 + Y^2))
            tmp = @. cis(q[1] * ((X - q[2])^2 + (Y - q[3])^2))
            -mapreduce(u -> overlap_integral(tmp, u), +, eachslice(uc; dims=3))
        end
        
        q = Optim.minimizer(optimize(J, zeros(3), LBFGS(), autodiff=:forward))
        # compensate for phase of non-plane wave
        @. uc /= cis(q[1] * ((X - q[2])^2 + (Y - q[3])^2))
    end

    # perform divison by mean intensity before computing the mcf
    uc ./= mean_uc

    polar ? calculate_isotropic2d_correlation_function_ft(uc, nthreads=Threads.nthreads()) : (1 // 2 .* (calculate_correlation_function_ft(uc; dim=1) .+ calculate_correlation_function_ft(uc; dim=2)))
end

function estimate_r0(mcf, D; llim=1e-6, ulim=1e2)
    mcf_pl(r0) = @. exp(-1 // 2 * 0.49 * r0^(-5 // 3) * D)
    J(r0) = norm(mcf_pl(r0[1]) .- mcf) # or overlap integral?: -overlap_integral(mcf_pl(r0[1]), mcf)
    r0_est = Optim.minimizer(optimize(J, llim, ulim)) |> first
    fit = mcf_pl(r0_est)

    r0_est, fit
end

function estimate_ccdf(v)
    # local ccdf
    cdffun = ecdf(v)

    l, u = extrema(v)

    # rice rule
    k = ceil(Int, 2cbrt(length(v)))
    x = range(l, u, length=k)

    ccdf = @. 1 - cdffun(x)

    x, ccdf
end

function estimate_pdf(v)

    l, u = extrema(v)
    # rice rule
    n = length(v)
    k = ceil(Int, 2cbrt(n))
    x = range(l, u, length=k)

    pdf = ash(v, rng=x, kernel=Kernels.epanechnikov).density

    x, pdf
end