using FINUFFT
using FFTW
using StatsBase
using Random
import FourierTools: fftshift!

include("helpers.jl")
include("ndgrid.jl")
include("phase_spectrum.jl")


logrange(x1, x2, n) = (sign(y) * (exp(abs(y)) - 1) for y in range(sign(x1) * log(abs(x1) + 1), sign(x2) * log(abs(x2) + 1), length=n))
asinhrange(x1, x2, n) = (sinh(y) for y in range(asinh(x1), asinh(x2), length=n))
sqrtrange(x1, x2, n) = (sign(y) * y^2 for y in range(sign(x1) * sqrt(abs(x1)), sign(x2) * sqrt(abs(x2)), length=n))


shift_screen!(dest::AbstractArray{T,N}, src::AbstractArray{Complex{T},N}; centered=false) where {T<:Real,N} = centered ? fftshift!(dest, real(src), 1:2) : copy!(dest, real(src))
shift_screen!(dest::AbstractArray{T,N}, src::AbstractArray{T,N}; centered=false) where {T<:Number,N} = centered ? fftshift!(dest, src, 1:2) : copy!(dest, src)

function uniform_phase_correction!(dest::AbstractArray{T}, dx, dy, δκx, δκy) where {T<:Complex}
    Nx, Ny = size(dest)
    x = fftfreq(Nx, Nx * dx)
    y = fftfreq(Ny, Ny * dy)
    @inbounds for I ∈ CartesianIndices(dest)
        m, n = I.I
        dest[I] *= cis(x[m] * δκx + y[n] * δκy)
    end
end

function demean_screen!(screen::AbstractArray{T,N}) where {T,N}
    @inbounds for I ∈ CartesianIndices(((axes(screen, d) for d ∈ 3:N)...,))
        m = mean(view(screen, :, :, I))
        for ℓ ∈ axes(screen, 2)
            @simd for k ∈ axes(screen, 1)
                screen[k, ℓ, I] -= m
            end
        end
    end
end


"""
    spectral_coefficients!(...)

    Cartesian grid, i.e., two distinct frequency vectors
"""
function spectral_coefficients!(c::AbstractArray{Complex{T},N}, Φ::Function, κx::AbstractVector, κy::AbstractVector, dκx::Number, dκy::Number; δκx=zero(T), δκy=zero(T)) where {T<:Real,N}

    randn!(c)

    @inbounds @simd for I ∈ CartesianIndices(c)
        m, n = I.I
        κ = sqrt((κx[m] + δκx)^2 + (κy[n] + δκy)^2)
        # factor two for correct variance in real and imaginary parts
        c[I] *= sqrt(2Φ(κ) * dκx * dκy)
    end
end
function spectral_coefficients!(c::AbstractArray{Complex{T},N}, Φ::Function, κx::AbstractVector, κy::AbstractVector, dκx::AbstractVector, dκy::AbstractVector; δκx=zero(T), δκy=zero(T)) where {T<:Real,N}

    randn!(c)

    @inbounds @simd for I ∈ CartesianIndices(c)
        m, n = I.I
        κ = sqrt((κx[m] + δκx)^2 + (κy[n] + δκy)^2)
        # factor two for correct variance in real and imaginary parts
        c[I] *= sqrt(2Φ(κ) * dκx[m] * dκy[n])
    end
end

"""
    spectral_coefficients!(...)

    Polar grid, i.e., one distinct frequency vector
"""
function spectral_coefficients!(c::AbstractArray{Complex{T},N}, Φ::Function, κ::AbstractVector, dκ::AbstractVector, dϕ::Number; δκ=zero(T)) where {T<:Real,N}

    randn!(c)

    @inbounds for I ∈ CartesianIndices(((axes(c, d) for d ∈ 3:N)...,))
        for n ∈ axes(c, 2)
            σ = sqrt(2Φ(κ[n] + δκ) * (κ[n] + δκ) * dκ[n] * dϕ)
            @simd for m ∈ axes(c, 1)
                # factor two for correct variance in real and imaginary parts
                c[m, n, I] *= σ
            end
        end
    end
end

function add_subharmonic!(dest::AbstractArray{T}, cn, x, y, κx, κy) where {T<:Complex}
    @inbounds @simd for I ∈ CartesianIndices(dest)
        m, n = I.I
        dest[I] += (cn * cis(κx * x[m] + κy * y[n]))
    end
end



function ft_phase_screen(; r0=0.1, Nx=256, dx=2 / Nx, dy=dx, l0=1e-2, L0=1e2, both=false, kwargs...)

    screen = Matrix{ComplexF64}(undef, Nx, Nx)
    ft_phase_screen!(screen; r0, dx, dy, l0, L0, kwargs...)

    both ? reim(screen) : real(screen)
end
"""
    ft_phase_screen!
    
    adapted from Schmidt 2010 - Numerical Simulation of Optical Wave Propagation With Examples in MATLAB
    random offset approach adapted from Paulson (2019), Journal of the Optical Society of America B, Vol. 36
"""
function ft_phase_screen!(screen::AbstractArray{Complex{T},N}; r0=T(0.1), dx=T(2 / size(screen, 1)), dy=dx, l0=T(1e-2), L0=T(1e2), Nsh=0, q=3, fft_plan=nothing, centered=true, pwd=false, model=andrews) where {T<:Union{Float32,Float64},N}

    # setup the spectrum
    Φ = phase_spectrum(r0, l0, L0, model)

    # preallocate complex Fourier coefficients
    Nx, Ny = size(screen)
    cn = zeros(Complex{T}, size(screen))

    # frequency grid [rad/m] 
    dκx = T(2π / (Nx * dx))
    dκy = T(2π / (Ny * dy))
    κx = fftfreq(Nx, Nx * dκx)
    κy = fftfreq(Ny, Ny * dκy)

    δκx = zero(T)
    δκy = zero(T)
    if pwd
        # save state of rng
        rng_state = copy(Random.default_rng())
        # random, uniform frequency offsets; generally speaking, there remains no DC component anymore! Use a different rng to simplify the comparison with the results from the other methods, i.e., make sure that the cn are always identical
        δκx = (rand(T) - T(0.5)) * dκx
        δκy = (rand(T) - T(0.5)) * dκy
        # restore state of rng
        copy!(Random.default_rng(), rng_state)
    end

    # "Fourier" coefficients
    spectral_coefficients!(cn, Φ, κx, κy, dκx, dκy; δκx, δκy)

    # set DC component to zero: "Because  the  zero  frequency  component  of  the phase-screen simulation alters only the total phase of the simulation, it does not change the spatial statistics of the fields, and therefore it is customary to set the zero frequency component to zeros" (Frehlich 2000). In case of random offsets, there is no DC component left  
    !(pwd && iszero(Nsh)) && for I ∈ CartesianIndices(((axes(cn, d) for d ∈ 3:N)...,))
        cn[1, 1, I.I...] = zero(Complex{T})
    end

    # synthesize the phase screen by superimposing the involved harmonics with an inverse DFT
    screenshifted = isnothing(fft_plan) ? bfft!(cn, 1:2) : fft_plan * cn

    # apply phase shift in spatial domain to compensate frequency shift in Fourier domain
    pwd && uniform_phase_correction!(screenshifted, dx, dy, δκx, δκy)

    shift_screen!(screen, screenshifted; centered)

    # now, add subharmonics if we are supposed to do so
    if Nsh > 0
        # preallocate spatial grid arrays
        x = fftfreq(Nx, Nx * dx) |> (centered ? fftshift : identity)
        y = fftfreq(Ny, Ny * dy) |> (centered ? fftshift : identity)

        # loop over frequency grids with spacing 1/(q^p*L)
        for p ∈ Base.OneTo(Nsh)

            # subharmonic frequency grid [rad/m] 
            dκx = T(2π / (Nx * dx) / (q^p))
            dκy = T(2π / (Ny * dy) / (q^p))
            κshx = fftfreq(q, q * dκx)
            κshy = fftfreq(q, q * dκy)

            δκx = zero(T)
            δκy = zero(T)

            for κy ∈ κshy
                for κx ∈ κshx
                    κ = sqrt(κx^2 + κy^2)
                    iszero(κ) && continue

                    if pwd
                        # save state of rng
                        rng_state = copy(Random.default_rng())
                        δκx = (rand(T) - T(0.5)) * dκx
                        δκy = (rand(T) - T(0.5)) * dκy
                        # restore state of rng
                        copy!(Random.default_rng(), rng_state)
                    end

                    # update κ to contain the random offsets needed for the pwd method
                    κ = sqrt((κx + δκx)^2 + (κy + δκy)^2)
                    let cn
                        cn = randn(Complex{T}) * sqrt(2Φ(κ) * dκx * dκy)
                        add_subharmonic!(screen, cn, x, y, κx + δκx, κy + δκy)
                    end
                end
            end
        end

        if pwd # the final low-frequency contribution 
            dκx = T(2π / (Nx * dx) / (q^(Nsh + 1)))
            dκy = T(2π / (Ny * dy) / (q^(Nsh + 1)))
            δκx = (rand(T) - T(0.5)) * dκx
            δκy = (rand(T) - T(0.5)) * dκy
            κ = sqrt(δκx^2 + δκy^2)
            let cn
                cn = randn(Complex{T}) * sqrt(2Φ(κ) * dκx * dκy)
                add_subharmonic!(screen, cn, x, y, δκx, δκy)
            end
        end

        demean_screen!(screen)
    end

    screen
end

function nu_spectral_grid(N, dx::T; sym=true, fun="asinh") where {T<:Real}

    κmax = π / dx

    rngfun = if fun == "asinh"
        asinhrange
    elseif fun == "log"
        logrange
    elseif fun == "sqrt"
        sqrtrange
    else
        error("$fun sampling not implemented")
    end

    # if the rngfuns would yield iterators(/mappedarrays?), this allocation could be avoided
    κ0 = collect(sym ? rngfun(-κmax, κmax, N + 1) : rngfun(zero(T), κmax, N + 1))

    dκ = [κ0[k+1] - κ0[k] for k ∈ Base.OneTo(N)]
    κ = [(κ0[k+1] + κ0[k]) / 2 for k ∈ Base.OneTo(N)]

    κ, dκ
end


function non_uniform_phase_screen(; r0=0.1, Nx=256, dx=2 / Nx, dy=dx, l0=1e-2, L0=1e2, both::Bool=false, kwargs...)

    screen = Matrix{ComplexF64}(undef, Nx, Nx)
    non_uniform_phase_screen!(screen; r0, dx, dy, l0, L0, kwargs...)

    both ? reim(screen) : real(screen)
end


function non_uniform_phase_screen!(screen::AbstractArray{Complex{T},N}; r0=0.1, dx=2 / size(screen, 1), dy=dx, l0=1e-2, L0=1e2, Nκx=size(screen, 1), Nκy=size(screen, 2), centered::Bool=true, tol=1e-7, fun="asinh", nthreads=FFTW.get_num_threads(), fftwflags=FFTW.ESTIMATE, upsampfac=0.0, model=andrews) where {T<:Union{Float32,Float64},N}

    # setup the spectrum
    Φ = phase_spectrum(r0, l0, L0, model)

    # preallocate complex fourier coefficients
    cn = zeros(Complex{T}, Nκx, Nκy, (size(screen, d) for d ∈ 3:N)...)

    # preallocate frequency grid
    κx, dκx = nu_spectral_grid(Nκx, T(dx); sym=true, fun)
    κy, dκy = nu_spectral_grid(Nκy, T(dy); sym=true, fun)
    spectral_coefficients!(cn, Φ, κx, κy, dκx, dκy)

    # CMCL mode ordering (neg. to pos.) ⟹ 0 OR FFTW mode ordering ⟹ 1
    modeord = Int(!centered)
    # synthesize the phase screen, frequencies in range [-3π,3π] required ⟹ normalization
    nufft2d1!(ndgrid(κx .* T(dx), κy .* T(dy))..., cn, 1, T(tol), screen; fftw=fftwflags, nthreads, upsampfac, modeord)

    demean_screen!(screen)

    screen
end

function non_uniform_phase_screen_pol(; r0=0.1, Nx=256, dx=2 / Nx, l0=1e-2, L0=1e2, both::Bool=false, kwargs...)

    screen = Matrix{ComplexF64}(undef, Nx, Nx)
    non_uniform_phase_screen_pol!(screen; r0, dx, l0, L0, kwargs...)

    both ? reim(screen) : real(screen)
end

function non_uniform_phase_screen_pol!(screen::AbstractArray{Complex{T},N}; r0=0.1, dx=2 / size(screen, 1), l0=1e-2, L0=1e2, Nκ=size(screen, 1), Nϕ=size(screen, 1), centered::Bool=true, tol=1e-7, fun="asinh", nthreads=FFTW.get_num_threads(), fftwflags=FFTW.ESTIMATE, upsampfac=0.0, model=andrews) where {T<:Union{Float32,Float64},N}

    # setup the spectrum
    Φ = phase_spectrum(r0, l0, L0, model)

    # preallocate complex fourier coefficients
    cn = zeros(complex(T), Nκ, Nκ, (size(screen, d) for d ∈ 3:N)...)

    # preallocate frequency grid
    κ, dκ = nu_spectral_grid(Nκ, T(dx); sym=false, fun)
    ϕ = collect(range(T(0), length=Nϕ))
    dϕ = T(2π / Nϕ)
    spectral_coefficients!(cn, Φ, κ, dκ, dϕ)

    # synthesize the phase screen
    κ .*= dx # frequencies in range [-3π,3π] required ⟹ normalization
    ϕ .*= dϕ

    # CMCL mode ordering (neg. to pos.) ⟹ 0 OR FFTW mode ordering ⟹ 1
    modeord = centered ? 0 : 1

    grid = ndgrid(ϕ, κ)
    pol2cart!(grid...)
    nufft2d1!(grid..., cn, 1, T(tol), screen; fftw=fftwflags, nthreads, upsampfac, modeord)

    demean_screen!(screen)

    screen
end