_f1d(arg,k,I) = [arg[i[k]] for i in I]


"""
generate_hg(x, y, z, l, m, w0, λ)

Calculate distribution of the HG_{l,m} beam mode at positions given by x, y, and z, where l is the horizontal (x) and m the vertical (y) mode index (both integers or zero). w0 is the beam radius of the corresponding fundamental Gaussian beam and λ the vacuum wavelength.
"""
function generate_hg(x, y, z, l=0, m=0, w0=1e-2, λ=1.55e-6)
    
    # l and m have to be ≥ 0
    @assert l ≥ 0
    @assert m ≥ 0

    hg = generate_hg_fun(l,m,w0,λ)

    ### parse inputs
    args = (x, y, z)
    nd = map(ndims, args)
    snd = sum(nd)

     # 1D
    if mapreduce(x -> x ≤ 1, &, nd)    
        I = CartesianIndices(map(length,args))
        args = ntuple(k -> _f1d(args[k],k,I), Val(3))

    # 2D
    elseif (4 ≤ snd ≤ 5) && mapreduce(x -> x ≤ 2, &, nd)

        # do nothing for one scalar element

        if snd == 5
            # one vector and two matrices
            vidx = filter(k -> (nd[k] == 1), 1:3)
            args = ntuple(k -> (k == vidx...) ? reshape(args[k], 1, 1, :) : args[k], Val(3))
        end
        
        # 3D
    elseif mapreduce(x -> x == 3, &, nd)
                
    else
        
        error("Dimensions of arguments are not compatible.")
        
    end
    
    u = broadcast(hg, args...)
    
    # find and drop singleton dimensions
    nsidx = filter(k -> k > 1, size(u))
    return reshape(u, nsidx)
end




"""
generate_hg_fun(l, m, w0, λ)

Generate anonymous function for HG_{l,m} beam mode, where l is the horizontal (x) and m the vertical (y) mode index (both integers, m,n >= 0). w0 is the beam radius of the corresponding fundamental Gaussian beam and λ the vacuum wavelength.
"""
function generate_hg_fun(l, m, w0, λ)
    k0 = 2π/λ
    #  rayleight length
    zr = k0*w0^2/2

    # 1/e beam radius
    w(z) = w0 * sqrt(1 + (z/zr)^2)
    #  radius of curvature of phase front
    R(z) = z + zr^2/z
    #  Gouy phase
    φ(z) = (1 + l + m) * atan(z,zr)


    # HG modes with propagator
    (x,y,z) -> w0 / w(z) * hermite(l, sqrt(2)*x/w(z)) * hermite(m, sqrt(2)*y/w(z)) * exp(-(x^2 + y^2)/w(z)^2) * cis(-k0*(x^2 + y^2)/2R(z) + φ(z) - k0*z)
end

"""
hermite(n, x)

Compute the value of the Hermite polynomial H_n(x). For further information see NIST Handbook, Olver 2010, p.443
"""
function hermite(n, x)
    y = zero(x)
    for ℓ ∈ 0:fld(n,2)
        y += (-one(x))^ℓ * (2x)^(n-2ℓ)/(factorial(ℓ)*factorial(n-2ℓ))
    end
    return y * factorial(n)
end