Base.@irrational sqrt2  1.4142135623730950488  √big(2)

_f1d(arg, k, I) = [arg[i[k]] for i in I]


"""
generate_lg(r, ϕ, z, l, m, w0, λ)

Calculate distribution of the LG_{l,m} beam mode at positions given by r, ϕ, and z, where l is the azimuthal and m the radial mode index (both integers, m >= 0). w0 is the beam radius of the corresponding fundamental Gaussian beam and λ the vacuum wavelength.
"""
function generate_lg(r, ϕ, z, l=0, m=0, w0=1e-2, λ=1.55e-6)

    # m has to be ≥ 0
    @assert m ≥ 0

    lg = generate_lg_fun(l, m, w0, λ)

    ### parse inputs
    args = (r, ϕ, z)
    nd = map(ndims, args)
    snd = sum(nd)

    # 1D inputs do not make sense for polar coordinates
    # 2D
    if (4 ≤ snd ≤ 5) && mapreduce(x -> x ≤ 2, &, nd)

        # do nothing for one scalar element

        if snd == 5
            # one vector and two matrices
            vidx = filter(k -> (nd[k] == 1), 1:3)
            args = ntuple(k -> (k == vidx...) ? reshape(args[k], 1, 1, :) : args[k], Val(3))
        end

        # 3D
    elseif mapreduce(x -> x == 3, &, nd)

    else

        error("Dimensions of arguments are not compatible.")

    end

    u = broadcast(lg, args...)

    # find and drop singleton dimensions
    nsidx = filter(k -> k > 1, size(u))
    return reshape(u, nsidx)
end


"""
generate_lg_cart(x, y, z, l, m, w0, λ)

Calculate distribution of the LG_{l,m} beam mode at positions given by x, y, and z, where l is the azimuthal and m the radial mode index (both integers, m >= 0). w0 is the beam radius of the corresponding fundamental Gaussian beam and λ the vacuum wavelength.
"""
function generate_lg_cart(x, y, z, l=0, m=0, w0=1e-2, λ=1.55e-6)

    # m has to be ≥ 0
    @assert m ≥ 0

    lg = generate_lg_fun_cart(l, m, w0, λ)

    ### parse inputs
    args = (x, y, z)
    nd = map(ndims, args)
    snd = sum(nd)

    # 1D
    if mapreduce(x -> x ≤ 1, &, nd)

        I = CartesianIndices(map(length, args))
        args = ntuple(k -> _f1d(args[k], k, I), Val(3))

        # 2D
    elseif (4 ≤ snd ≤ 5) && mapreduce(x -> x ≤ 2, &, nd)

        # do nothing for one scalar element

        if snd == 5
            # one vector and two matrices
            vidx = filter(k -> (nd[k] == 1), 1:3)
            args = ntuple(k -> (k == vidx...) ? reshape(args[k], 1, 1, :) : args[k], Val(3))
        end

        # 3D
    elseif mapreduce(x -> x == 3, &, nd)

    else
        error("Dimensions of arguments are not compatible.")
    end

    u = broadcast(lg, args...)

    # find and drop singleton dimensions
    nsidx = filter(k -> k > 1, size(u))
    return reshape(u, nsidx)
end



"""
generate_lg_fun(l, m, w0, λ)

Generate anonymous function for LG_{l,m} beam mode, where l is the azimuthal and m the radial mode index (both integers, m >= 0). w0 is the beam radius of the corresponding fundamental Gaussian beam and λ the vacuum wavelength.
"""
function generate_lg_fun(l, m, w0, λ)
    k0 = 2(π / λ)
    #  rayleigh length
    zr = k0 * w0^2 / 2

    # 1/e beam radius
    w(z) = w0 * sqrt(1 + (z / zr)^2)
    #  radius of curvature of phase front
    R(z) = z + zr^2 / z
    #  Gouy phase
    φ(z) = (1 + 2m + abs(l)) * atan(z, zr)


    # LG modes with propagator
    (r, ϕ, z) -> w0 / w(z) * (sqrt2 * r / w(z))^abs(l) * exp(-(r / w(z))^2) * generalized_laguerre(m, abs(l), 2 * (r / w(z))^2) * cis(-(l * ϕ + k0 * r^2 / (2R(z)) - φ(z) - k0 * z))
end
function generate_lg_fun_cart(l, m, w0, λ)
    r(x, y) = sqrt(x^2 + y^2)
    ϕ(x, y) = atan(y, x)

    lg = generate_lg_fun(l, m, w0, λ)

    return (x, y, z) -> lg(r(x, y), ϕ(x, y), z)
end


"""
generalized_laguerre(n, α, x)

Compute the value of the generalized Laguerre polynomial L_n^α(x). For further information see NIST Handbook, Olver 2010, p.443
"""
function generalized_laguerre(n, α, x::T) where T
    y = zero(T)
    for l in 0:n
        y += (-1)^l * factorial(n + α) // (factorial(n - l) * factorial(α + l) * factorial(l)) * x^l
    end
    return y
end