using Bessels
using Roots

"""
generate_lp(r, ϕ, l, m)

Calculate distribution of the LP_{l,m} beam mode at positions given by r and ϕ, where l is the azimuthal and m the radial mode index (both integers, m > 0). even_odd determines whether the azimuthal dependency follows a sine (odd) or a cosine (even) pattern. a is the fiber core radius and V the normalized frequency.
"""
function generate_lp!(even::AbstractArray{T}, r::AbstractArray{R}, ϕ::AbstractArray{R}; l::Integer=0, m::Integer=1, a=4.5e-6, V=1.7) where {T<:Union{R,Complex{R}}} where {R<:Real}
    lpfun = generate_lp_fun(l, m, a, V)
    map!(lpfun, even, r, ϕ)
end
function generate_lp!(even::AbstractArray{T}, odd::AbstractArray{T}, r::AbstractArray{R}, ϕ::AbstractArray{R}; l::Integer=0, m::Integer=1, a=4.5e-6, V=1.7) where {T<:Union{R,Complex{R}}} where {R<:Real}
    lpfun = generate_lp_fun(l, m, a, V)
    map!((r, ϕ) -> lpfun(r, ϕ, 0), even, r, ϕ), map!((r, ϕ) -> lpfun(r, ϕ, π / 2), odd, r, ϕ)
end
function generate_lp(r::AbstractArray{T}, ϕ::AbstractArray{T}; l::Integer=0, m::Integer=1, a=4.5e-6, V=1.7) where {T<:Real}
    if iszero(l)
        generate_lp!(zero(r), r, ϕ; l, m, a, V)
    else
        generate_lp!(zero(r), zero(r), r, ϕ; l, m, a, V)
    end
end
function generate_lp(r::T, ϕ::T; l::Integer=0, m::Integer=1, a=4.5e-6, V=1.7) where {T<:Real}
    lpfun = generate_lp_fun(l, m, a, V)
    if iszero(l)
        lpfun(r, ϕ, 0)
    else
        lpfun(r, ϕ, 0), lpfun(r, ϕ, π / 2)
    end
end

"""
generate_lp_fun(l, m[, a, V])

Generate anonymous function for LP_{l,m} beam mode, where l is the azimuthal and m the radial mode index (both integers, m > 0). a is the fiber core radius and V the normalized frequency.
"""
function generate_lp_fun(l::Integer, m::Integer, a=4.5e-6, V=1.7)

    # obtain normalized transverse propagation constant from dispersion relation
    u = solve_dispersion_relation(l, m, V)
    w = sqrt(V^2 - u^2)

    (r, ϕ, ψ=0) -> (r ≤ a) ? besselj(l, u * r / a) * cos(l * ϕ + ψ) : besselj(l, u) / besselk(l, w) * besselk(l, w * r / a) * cos(l * ϕ + ψ)
end



"""
solve_dispersion_relation(l, m, V)

Solve dispersion relation for LP_{l,m} in fiber with normalized frequency V. l is the azimuthal and m the radial mode index.
"""
function solve_dispersion_relation(l::Integer, m::Integer, V::Real)

    ϵ = 1e-14

    w(u) = sqrt(V^2 - u^2)
    f(u) = u * besselj(l - 1, u) / besselj(l, u) + w(u) * besselk(l - 1, w(u)) / besselk(l, w(u))

    lower = isone(m) ? 0 : besselj_l_kth_zero(l, m - 1)
    lower += ϵ
    upper = min(besselj_l_kth_zero(l, m), V) - ϵ

    (lower >= V) && throw("The cut-off frequency of mode ($l,$m) Vc=$lower exceeds V=$V.")

    find_zero(f, (lower, upper))
end


"""
besselj_l_kth_zero(l, k)

Obtain kth zero of J_l(x)
"""
function besselj_l_kth_zero(l::Integer, k::Integer)

    # J_{-l} = (-1)^l * J_l
    l = abs(l)

    # see NIST Handbook, Olver 2010 (10.21 (vi))
    a = (k + l//2 - 1//4)*π
    µ = 4*l^2
    # this approximation holds for large zeros, but should provide a fairly accurate starting point 
    j_nu_k = a - (µ-1)/8/a - 4*(µ-1)*(7µ-31)/3/(8a)^3 - 32*(µ-1)*(83*µ^2 - 982µ + 3779)/15/(8a)^5 - 64*(µ-1)*(6949*µ^3 - 153855*µ^2 + 1585743µ - 6277237)/105/(8a)^7
    
    # refine
    find_zero(x -> besselj(l, x), j_nu_k)
end