include("generate_lg.jl")

struct Domain{T}
    # dimensions: transverse & longitudinal
    Dx::T
    Dy::T
    Dz::T
    # step sizes: transverse & longitudinal
    dx::T
    dy::T
    dz::T
    # samples: transverse & longitudinal
    Nx::Int
    Ny::Int
    Nz::Int
    # vectors: transverse & longitudinal
    x::AbstractVector{T}
    y::AbstractVector{T}
    z::AbstractVector{T}

    # circular aperture radius
    rA::T

    function Domain{T}(dx, dy, dz, Nx, Ny, Nz, rA) where {T<:Real}
        Dx = dx * Nx
        Dy = dy * Ny
        Dz = dz * (Nz - 1)

        x = fftfreq(Nx, Nx*dx) |> fftshift
        y = fftfreq(Ny, Ny*dy) |> fftshift
        z = (0:(Nz-1)) * dz

        new{T}(Dx, Dy, Dz, dx, dy, dz, Nx, Ny, Nz, x, y, z, rA)
    end
end
Domain(dx, dy, dz, Nx, Ny, Nz, rA) = Domain{Float64}(dx, dy, dz, Nx, Ny, Nz, rA)
Domain{T}(dxy, dz, Nxy, Nz, rA) where {T<:Real} = Domain{T}(dxy, dxy, dz, Nxy, Nxy, Nz, rA)
Domain(dxy, dz, Nxy, Nz, rA) = Domain{Float64}(dxy, dxy, dz, Nxy, Nxy, Nz, rA)


struct Beam{T}
    l::Int
    m::Int

    λ::T
    k0::T
    w0::T
    wlm::T
    M²::T

    fun::Function

    function Beam{T}(l, m, λ, w0) where {T<:Real}
        M² = abs(l) + 2m + 1
        new{T}(l, m, λ, 2π / λ, w0, w0 * sqrt(M²), M², generate_lg_fun_cart(l, m, T(w0), T(λ)))
    end
end
Beam(l, m, λ, w0) = Beam{Float64}(l, m, λ, w0)


# Far from optimal
struct FiberModes{T}
    a::T # core radius
    NA::T # numerical aperture
    MFR::T # mode field radius (fundamental mode)

    modes::OrderedDict{Tuple{Int,Int,Symbol},AbstractMatrix{Complex{T}}}

    # parameter which is not used for FiberModes
    FiberModes{T}(a, NA, MFR=NaN; kwargs...) where {T <: Real} = new{T}(a, NA, MFR, OrderedDict{Tuple{Int,Int,Symbol},AbstractMatrix{Complex{T}}}())
end
FiberModes(a, NA, MFR=NaN; kwargs...) = FiberModes{Float64}(a, NA, MFR; kwargs...)


struct FreeSpaceModes{T}
    w::T # beam waist of fundamental beam mode
    nmodes::Int # number of beam modes
    which::Symbol

    modes::OrderedDict{NTuple{2,Int},AbstractMatrix{Complex{T}}}

    FreeSpaceModes{T}(w, n; which=:HG) where {T} = new{T}(w, n, which, OrderedDict{NTuple{2,Int},AbstractMatrix{Complex{T}}}())
end
FreeSpaceModes(w, n; which=:HG) = FreeSpaceModes{Float64}(w, n; which)

const Modes = Union{FiberModes,FreeSpaceModes}
generate_wgs(f=FiberModes; which=:HG, args...) = NamedTuple(k => f(v...; which) for (k, v) ∈ pairs(args))