using LinearAlgebra
using FFTW
import FourierTools: fftshift!

function isuniform(x::AbstractVector{T}, ε=2eps(T)) where {T<:AbstractFloat}
    (length(x) <= 1) && return true
    !isfinite(last(x) - first(x)) && return false

    # Create uniform range as reference. If deviation from reference exceeds ϵ, spacing is nonuniform.
    n = length(x)
    xu = range(first(x), last(x), length=n)

    return all(i -> abs(x[i] - xu[i]) <= ε, Base.OneTo(n))
end

"""
freespace_tf_fr()

Compute the transfer function of free-space propagation under the Fresnel approximation.
"""
freespace_tf_fr(κ², k0, Δz) = cis(Δz * (k0 - κ² / 2k0))



"""
split_step_method!(u,λ,dx,z,R)

Propagate the input field distribution u with a wavelength λ, generated on a transverse grid with resolution dx, along the z direction where z is a vector containing the discrete z-positions. R is a vector of matrices representing the refraction operators.
"""
function split_step_method!(u::AbstractMatrix{Complex{T}}, λ, dx, dy, z, R::AbstractArray{Complex{T},3}) where {T<:Union{Float32,Float64}}

    # assume square, uniform grid
    M, N = size(u)

    # first element must correspond to 0 for FFT
    mg = fftfreq(M, one(T))
    ng = fftfreq(N, one(T))
    # Super-Gaussian window
    w = [exp(-((xn^2 + yn^2) / T(0.47)^2)^8) for xn ∈ mg, yn ∈ ng]

    # Vaccum wavenumber
    k0 = 2(π / λ)

    # preallocate angular spatial frequencies
    dκx = 2(π / (M * dx))
    dκy = 2(π / (N * dy))
    κx = fftfreq(M, M * dκx)
    κy = fftfreq(M, N * dκy)
    κ² = [κx^2 + κy^2 for κx ∈ κx, κy ∈ κy]
    # shifted field
    us = ifftshift(u)

    P = plan_fft!(us)
    iP = plan_ifft!(us)

    # transfer function
    Δz = (z[2] - z[1]) / 2
    H = [freespace_tf_fr(κ², k0, Δz) for κ² ∈ κ²]
    # uniform z axis
    uflag = isuniform(z)

    # SYMMETRIC SPLIT-STEP
    @inbounds for m ∈ 1:length(z)-1
        !uflag && (@. H = freespace_tf_fr(κ², k0, (z[m+1] - z[m]) / 2))

        # Windowing
        us .*= w
        # Diffraction
        ang_spec_prop!(us, H, P, iP)
        # Refraction
        us .*= w .* view(R, :, :, m)

        # Diffraction
        ang_spec_prop!(us, H, P, iP)
    end

    fftshift!(u, us)
    u
end


"""
ang_spec_prop!(u,U,H,P,iP)

Propagate the input field distribution u (shifted s.t. it is compatible with the FFT/IFFT). Apply the diffraction operator H and overwrite u. Replace u with its diffracted counterpart by overwriting u with the IFFT output (operator iP).
Corresponds to angular spectrum propagation.
"""
function ang_spec_prop!(u, H, P, iP)
    # FFT
    P * u
    # Diffraction
    @inbounds @simd for k ∈ eachindex(u)
        u[k] *= H[k]
    end
    # IFFT
    iP * u
end

"""
onestep_prop_ir!(u,dx1,k0,z)

Propagate the input field distribution u (shifted s.t. it is compatible with the FFT/IFFT). This function evaluates the Fresnel diffraction integral in a single step. Source and observation plane grid differ!
"""
function onestep_prop_ir!(u, dx1, k0, z)

    N = size(u, 1)
    # source plane grid
    xy1 = fftfreq(N, N * dx1)
    # observation plane grid
    dx2 = 2π * z / (k0 * N * dx1)
    xy2 = fftfreq(N, N * dx2)

    # apply quadratic phase in source plane
    @inbounds for m ∈ Base.OneTo(N)
        @simd for n ∈ Base.OneTo(N)
            r1² = xy1[n]^2 + xy1[m]^2
            u[n, m] *= cis(k0 * r1² / 2z)
        end
    end

    # FFT
    fft!(u)

    # apply quadratic phase in observation plane
    a = -1im * k0 / (2π * z)
    @inbounds for m in Base.OneTo(N)
        @simd for n in Base.OneTo(N)
            r2² = xy2[n]^2 + xy2[m]^2
            u[n, m] *= a * cis(k0 * r2² / 2z)
        end
    end

    return u, dx2
end


"""
twostep_prop_ir!(u,dx1,dx2,k0,z)

Propagate the input field distribution u (shifted s.t. it is compatible with the FFT/IFFT). This function makes use of the impulse-response-based two-step method and hence allows for different spatial resolutions in the source (dx1) and observation plane (dx2). Make sure that the sampling criterion is fulfilled (oversampling --> IR oversampled, TF undersampled). For further information on the function see chapter 6 in the book by Schmidt (2010). For further information on the sampling see chapter 5 in the the book by Voelz (2011).
"""
function twostep_prop_ir!(u, dx1, dx2, k0, Δz)

    N = size(u, 1)
    # source plane grid
    xy1 = fftfreq(N, N * dx1)
    # magnification
    mag = dx2 / dx1

    # intermediate plane grid
    Δz1 = Δz / (1 - mag)
    Δz2 = Δz - Δz1
    dxi = 2π * abs(Δz1) / (k0 * N * dx1)
    xyi = fftfreq(N, N * dxi)

    # observation plane grid
    xy2 = fftfreq(N, N * dx2)

    # apply quadratic phase in source plane
    @inbounds for m ∈ Base.OneTo(N)
        @simd for n ∈ Base.OneTo(N)
            r1² = xy1[n]^2 + xy1[m]^2
            u[n, m] *= cis(k0 / 2Δz1 * r1²)
        end
    end

    # FFT
    fft!(u)
    u .*= dx1^2

    # apply quadratic phase in intermediate plane
    a = -1im * k0 / (2π * Δz1)
    @inbounds for m in Base.OneTo(N)
        @simd for n in Base.OneTo(N)
            ri² = xyi[n]^2 + xyi[m]^2
            u[n, m] *= a * cis(k0 * Δz / (2Δz1 * Δz2) * ri²)
        end
    end

    # FFT
    fft!(u)
    u .*= dxi^2

    # apply quadratic phase in observation plane
    a = -1im * k0 / (2π * Δz2)
    @inbounds for m in Base.OneTo(N)
        @simd for n in Base.OneTo(N)
            r2² = xy2[n]^2 + xy2[m]^2
            u[n, m] *= a * cis(k0 / 2Δz2 * r2²)
        end
    end

    u
end


"""
twostep_prop_ang!(u,dx1,dx2,k0,z)

Propagate the input field distribution u (shifted s.t. it is compatible with the FFT/IFFT). This function makes use of the transfer-function-based (angular spectrum) two-step method and hence allows for different spatial resolutions in the source (dx1) and observation plane (dx2). Make sure that the sampling criterion is fulfilled (undersampling --> IR undersampled, TF oversampled). For further information on the function see chapter 6 in the book by Schmidt (2010). For further information on the sampling see chapter 5 in the the book by Voelz (2011).
"""
function twostep_prop_ang!(u, dx1, dx2, k0, Δz, κlim=1)

    N = size(u, 1)

    # source plane grid
    xy1 = fftfreq(N, N * dx1)

    # magnification
    mag = dx2 / dx1
    dκ = 2π / (N * dx1)
    κxy = fftfreq(N, N * dκ)

    # observation plane grid
    xy2 = fftfreq(N, N * dx2)

    # apply quadratic phase in source plane
    @inbounds for m ∈ Base.OneTo(N)
        @simd for n ∈ Base.OneTo(N)
            r1² = xy1[n]^2 + xy1[m]^2
            u[n, m] *= cis(k0 / 2Δz * (1 - mag) * r1²) / mag
        end
    end

    # FFT
    fft!(u)

    # apply quadratic phase in intermediate plane and filter
    κc² = 2(N * dκ / 2 * κlim)^2
    @inbounds for m in Base.OneTo(N)
        @simd for n in Base.OneTo(N)
            κ² = κxy[n]^2 + κxy[m]^2
            u[n, m] *= (κ² > κc²) ? zero(eltype(u)) : cis(-Δz / (2mag * k0) * κ²)
        end
    end
    # # apply quadratic phase in intermediate plane
    # @inbounds for m in Base.OneTo(N)
    #     @simd for n in Base.OneTo(N)
    #         κ² = κxy[n]^2 + κxy[m]^2
    #         u[n,m] *= cis(-Δz/(2mag*k0) * κ²)
    #     end
    # end

    # IFFT
    ifft!(u)

    # apply quadratic phase in observation plane
    @inbounds for m in Base.OneTo(N)
        @simd for n in Base.OneTo(N)
            r2² = xy2[n]^2 + xy2[m]^2
            u[n, m] *= cis(k0 / 2Δz * (1 - 1 / mag) * r2²)
        end
    end

    u
end
