using FINUFFT # SEGFAULT may occur if FINUFFT is not loaded sufficiently early
using StatsBase
using AverageShiftedHistograms
using ProgressMeter
using FFTW
using HDF5
using H5Zblosc
using Random
using OrderedCollections

include("newtypes.jl")
include("ft_phase_screen.jl")
include("generate_lg.jl")
include("freespace_coupling.jl")
include("split_step_method.jl")
include("estimate_statistics.jl")
include("plot_basics.jl")


function screen_to_refraction_op(screen::AbstractArray{T,3}) where {T<:Complex}
    R = Array{T,3}(undef, size(screen, 1), size(screen, 2), 2size(screen, 3))

    @inbounds for m ∈ axes(screen, 3), ℓ ∈ axes(screen, 2), k ∈ axes(screen, 1)
        R[k, ℓ, 2m-1] = cis(real(screen[k, ℓ, m]))
        R[k, ℓ, 2m] = cis(imag(screen[k, ℓ, m]))
    end

    R
end


function turbulent_channel_model(; domain::Domain{T}=Domain(0.3/256, 1e2, 256, 11, 1e-2), beam::Beam{T}=Beam(0, 0, 1.55e-6, 1e-2), Cn²=1e-14, l0=1e-3, L0=1e1, uniform::Bool=false, model=andrews, pw::Bool=false) where {T<:Union{Float32,Float64}}
    
    # generate plane wave or Gaussian beam wave
    u = pw ? ones(Complex{T}, domain.Nx, domain.Ny) : broadcast(beam.fun, domain.x, domain.y |> transpose, zero(T))

    # fried parameter
    r0 = T((0.423 * beam.k0^2 * domain.dz * Cn²)^(-3 / 5))

    # choose number of subharmonics Nsh such that 3^Nsh * domain.d ≥ L0
    Nsh = round(Int, cld(log(L0 / min(domain.Dx, domain.Dy)), log(3))) # we assume L0 > d
    # we need Nz-1 screens since Nz takes the source plane into account
    Ncplx = cld(domain.Nz - 1, 2)

    # preallocate phase screen array
    screen = Array{Complex{T},3}(undef, size(u)..., Ncplx)

    # Uniform sampling with subharmonics vs nonuniformly distributed plane waves
    uniform ? ft_phase_screen!(screen; r0, dx=domain.dx, dy=domain.dy, l0, L0, centered=false, Nsh, model) :
    non_uniform_phase_screen!(screen; r0, dx=domain.dx, dy=domain.dy, l0, L0, centered=false, nthreads=1, tol=1e-5, model)

    R = screen_to_refraction_op(screen)
    split_step_method!(u, beam.λ, domain.dx, domain.dy, domain.z, R)
end
